
#ifndef GLOBAL_H
#define GLOBAL_H


#include <map>
#include "buffer.h" // for |id_type|
#include "axis-types.h" // for |shared_value|
#include "parse_types.h" 

#include <string>
#include <sstream>

#include "arithmetic.h"
#include "bigint.h"

#include "../Atlas.h" // type declarations that are ``common knowledge''
#include "matrix.h" // to make |int_Vector| and |int_Matrix| complete types
#include "ratvec.h" 

namespace atlas { namespace interpreter {

typedef std::shared_ptr<shared_value> shared_share;
class id_data
{ shared_share val;  type_expr tp;  bool is_constant;
public:
  id_data(shared_share&& val,type_expr&& t,bool is_const)
  : val(std::move(val)), tp(std::move(t)), is_constant(is_const) {}
#ifdef incompletecpp11
  id_data () : val(), tp(), is_constant(false) {}
    // we \emph{must} have a default constructor
  id_data (const id_data& x) : val(), tp(), is_constant(false) {}
   // and allow copying such value
  id_data& operator= (const id_data& x) = delete;
  id_data (id_data&& x)
  : val(std::move(x.val)), tp(std::move(x.tp)), is_constant(x.is_constant){}
  id_data& operator=(id_data&& x)
  { val = std::move(x.val); tp = std::move(x.tp); is_constant = x.is_constant;
    return *this;
  }
#else
  id_data (id_data&& x) = default;
  id_data& operator=(id_data&& x) = default; // no copy-and-swap needed
#endif
  void swap(id_data& x)  { val.swap(x.val); tp.swap(x.tp); }

  const shared_share& value() const { return val; }
  const type_expr& type() const { return tp; }
  type_expr& type() { return tp; }
  // non-|const| reference; may be specialised by caller
  bool is_const() const { return is_constant; }
};

class Id_table
{ typedef std::map<id_type,id_data> map_type;
  map_type table;
public:
  Id_table(const Id_table&) =  delete ;
  Id_table& operator=(const Id_table&) =  delete ;
  Id_table() : table() {} // the default and only constructor

  void add(id_type id, shared_value v, type_expr&& t, bool is_const);
   // insertion
  void add_type_def(id_type id, type_expr&& t); // insertion of type only
  bool remove(id_type id); // deletion
  shared_share address_of(id_type id); // locate

  bool present (id_type id) const
  { return table.find(id)!=table.end(); }
  bool is_defined_type(id_type id) const; // whether |id| stands for a type
  const_type_p type_of(id_type id,bool& is_const) const;
  // pure lookup, may return |nullptr|
  void specialise(id_type id,const type_expr& type);
  // specialise type stored for identifier
  shared_value value_of(id_type id) const; // look up

  size_t size() const { return table.size(); }
  void print(std::ostream&) const;
};

class function_base;
// derived from |value_base|, defined in \.{axis.h}; values with function type
typedef std::shared_ptr<const function_base> shared_function;
// specialises |shared_value|

class overload_data
{ shared_function val;  func_type tp;
public:
  overload_data(shared_function&& val,func_type&& t)
  : val(std::move(val)), tp(std::move(t)) {}
#ifdef incompletecpp11
  overload_data (const overload_data& x) = delete;
  overload_data& operator=(const overload_data& x) = delete;
  overload_data(overload_data&& x)
  : val(std::move(x.val)), tp(std::move(x.tp)) {}
  overload_data& operator=(overload_data&& x)
  { val = std::move(x.val); tp = std::move(x.tp); return *this; }
#else
  overload_data (overload_data&& x) = default;
  overload_data& operator=(overload_data&& x)
   = default; // no copy-and-swap needed
#endif

  const shared_function& value() const { return val; }
  const func_type& type() const { return tp; }
};

class overload_table
{
public:
#ifdef incompletecpp11
  class variant_list : public std::vector<overload_data>
  { typedef std::vector<overload_data> Base;
  public:
    variant_list() : Base()  {}
    variant_list (const variant_list& x)
      // \emph{required} copy constructor, for empty vectors only
    : Base() { assert(x.size()==0); }
    variant_list (variant_list&& x) : Base(std::move(x)) {}
    variant_list& operator=(const variant_list& x)=delete;
  };
#else
  typedef std::vector<overload_data> variant_list;
#endif
  typedef std::map<id_type,variant_list> map_type;
private:
  map_type table;
public:
  overload_table (const Id_table&) =delete;
  overload_table& operator=(const Id_table&) = delete;
  overload_table() : table() {} // the default and only constructor
 // accessors
  const variant_list& variants(id_type id) const;
  const overload_data* entry (id_type id, const type_expr& arg_t) const;
  size_t size() const { return table.size(); }
   // number of distinct identifiers
  void print(std::ostream&) const;
 // manipulators
  void add(id_type id, shared_function v, type_expr&& t);
   // insertion
  bool remove(id_type id, const type_expr& arg_t); //deletion
};

class type_table
{ public:  typedef std::vector<id_type> field_list;
  static constexpr id_type no_id = -1; // value to signal absence of field name

  private: std::vector<std::pair<id_type,field_list> > assoc;
public:
  type_table() : assoc()  {}
  void add (id_type id, field_list&& names);
  void remove (id_type id);

  id_type find (const type_expr& type) const;
    // identifier defined as |type|, or |no_id|
  const field_list& fields (id_type type) const;
};

class definition_group
{ typedef std::vector<std::pair<id_type,type_expr> > association;
  association bindings;
  BitMap constness;
public:
  definition_group(unsigned int n_ids);
 // manipulators
  void add(id_type id,type_expr&& t, bool is_const);
  void thread_bindings (const id_pat& pat,const type_expr& type);
  association::iterator begin() { return bindings.begin(); }
  association::iterator end() { return bindings.end(); }
 // accessors
  bool is_const (association::const_iterator it) const
  { return constness.isMember(it-bindings.cbegin()); }
};

struct int_value : public value_base
{ arithmetic::big_int val;

  explicit int_value(int v) : val(v)  {}
  explicit int_value(arithmetic::Numer_t v)
  : val(big_int::from_signed(v))  {}
  explicit int_value(unsigned int v)
    : val(big_int::from_unsigned(v))  {}
  explicit int_value(unsigned long v)
    : val(big_int::from_unsigned(v))  {}
  explicit int_value(unsigned long long v)
    : val(big_int::from_unsigned(v))  {}
  explicit int_value(arithmetic::big_int&& v) : val(std::move(v))  {}
  void print(std::ostream& out) const { out << val; }
  int_value* clone() const { return new int_value(*this); }
  static const char* name() { return "integer"; }

  int int_val () const { return val.int_val(); }
private:
  int_value(const int_value& v) : val(v.val) {}
};

typedef std::shared_ptr<const int_value> shared_int;
typedef std::shared_ptr<int_value> own_int;

struct rat_value : public value_base
{ big_rat val;

  explicit rat_value(Rational v) : val(v)  {}
  explicit rat_value(big_rat&& r) : val(std::move(r)) {}

  void print(std::ostream& out) const { out << val; }
  rat_value* clone() const { return new rat_value(*this); }
  static const char* name() { return "integer"; }

#ifdef incompletecpp11
  big_int numerator() const { return val.numerator(); }
  big_int denominator() const { return val.denominator(); }
  big_int& numerator() { return val.numerator(); }
  big_int& denominator() { return val.denominator(); }
#else
  big_int numerator() const & { return val.numerator(); }
  big_int denominator() const & { return val.denominator(); }
  big_int&& numerator() & { return std::move(val).numerator(); }
  big_int&& denominator() & { return std::move(val).denominator(); }
#endif
  Rational rat_val() const { return val.rat_val(); }

private:
  rat_value(const rat_value& v) : val(v.val) {}
};

typedef std::shared_ptr<const rat_value> shared_rat;
typedef std::shared_ptr<rat_value> own_rat;

struct string_value : public value_base
{ std::string val;

  explicit string_value(const std::string& s) : val(s)  {}
  explicit string_value(std::string&& s) : val(std::move(s))  {}
  template <typename I> string_value(I begin, I end) : val(begin,end)  {}
  ~string_value() {}
  void print(std::ostream& out) const { out << '"' << val << '"'; }
  string_value* clone() const { return new string_value(*this); }
  static const char* name() { return "string"; }
private:
  string_value(const string_value& v) : val(v.val) {}
};

typedef std::shared_ptr<const string_value> shared_string;
typedef std::shared_ptr<string_value> own_string;
 

struct bool_value : public value_base
{ bool val;

  explicit bool_value(bool v) : val(v)  {}
  ~bool_value() {}
  void print(std::ostream& out) const { out << std::boolalpha << val; }
  bool_value* clone() const { return new bool_value(*this); }
  static const char* name() { return "Boolean"; }
private:
  bool_value(const bool_value& v) : val(v.val) {}
};

typedef std::shared_ptr<const bool_value> shared_bool;

struct vector_value : public value_base
{ int_Vector val;

  explicit vector_value(const std::vector<int>& v) : val(v)  {}
  explicit vector_value(std::vector<int>&& v) : val(std::move(v))  {}
  template <typename I> vector_value(I begin, I end) : val(begin,end)  {}
  ~vector_value() {}
  virtual void print(std::ostream& out) const;
  vector_value* clone() const { return new vector_value(*this); }
  static const char* name() { return "vector"; }
private:
  vector_value(const vector_value& v) : val(v.val) {}
};

typedef std::shared_ptr<const vector_value> shared_vector;
typedef std::shared_ptr<vector_value> own_vector;

struct matrix_value : public value_base
{ int_Matrix val;

  explicit matrix_value(const int_Matrix& v) : val(v)  {}
  explicit matrix_value(int_Matrix&& v) : val(std::move(v))  {}
  ~matrix_value() {}
  virtual void print(std::ostream& out) const;
  matrix_value* clone() const { return new matrix_value(*this); }
  static const char* name() { return "matrix"; }
private:
  matrix_value(const matrix_value& v) : val(v.val) {}
};

typedef std::shared_ptr<const matrix_value> shared_matrix;
typedef std::shared_ptr<matrix_value> own_matrix;

struct rational_vector_value : public value_base
{ RatWeight val;

  explicit rational_vector_value(const RatWeight& v)
   : val(v) { val.normalize();}
  rational_vector_value(const int_Vector& v,int d)
   : val(v,d)  { val.normalize(); }
  rational_vector_value(matrix::Vector<arithmetic::Numer_t>&& v,
                       arithmetic::Denom_t d)
   : val(std::move(v),d)  { val.normalize(); }
  template <typename I>
     rational_vector_value(I begin, I end, arithmetic::Denom_t d)
    : val(matrix::Vector<arithmetic::Numer_t>(begin,end),d)
    { val.normalize(); }
  ~rational_vector_value() {}
  virtual void print(std::ostream& out) const;
  rational_vector_value* clone() const
   { return new rational_vector_value(*this); }
  static const char* name() { return "rational vector"; }
private:
  rational_vector_value(const rational_vector_value& v) : val(v.val) {}
};

typedef std::shared_ptr<const rational_vector_value> shared_rational_vector;
typedef std::shared_ptr<rational_vector_value> own_rational_vector;

typedef void (* wrapper_function)(expression_base::level);


extern Id_table* global_id_table;

extern overload_table* global_overload_table;

extern type_table typedef_table;

extern std::ostream* output_stream;

extern bool clean;

extern int verbosity;

extern const shared_bool global_false, global_true;


std::ostream& operator<< (std::ostream& out, const Id_table& p);

std::ostream& operator<< (std::ostream& out, const overload_table& p);

void initialise_evaluator();

type_expr analyse_types(const expr& e,expression_ptr& p);

void global_set_identifier(const struct raw_id_pat& id, expr_p e, int overload);
void global_set_identifiers(const raw_let_list& d);
void global_declare_identifier(id_type id, type_p type);
void global_forget_identifier(id_type id);
void global_forget_overload(id_type id, type_p type);
void type_define_identifier
  (id_type id, type_p type, raw_id_pat ip, const YYLTYPE& loc);
void show_ids(std::ostream& out);
void type_of_expr(expr_p e);
void show_overloads(id_type id,std::ostream& out);

void install_function
 (wrapper_function f,const char*name, const char* type_string);

void sizeof_vector_wrapper(expression_base::level l);
void sizeof_ratvec_wrapper(expression_base::level l);
void sizeof_string_wrapper(expression_base::level l);
void matrix_ncols_wrapper(expression_base::level l);
void virtual_module_size_wrapper(expression_base::level l);

void mm_prod_wrapper (expression_base::level);

void transpose_mat_wrapper (expression_base::level);
void id_mat_wrapper(expression_base::level l);


template <typename T>
  std::string str(T n) { std::ostringstream s; s<<n; return s.str(); }
inline std::string str(unsigned char c)
  { return str(static_cast<unsigned int>(c)); }

inline shared_bool whether(bool b){ return b ? global_true : global_false; }

} }
#endif

